;; Generated package description from triples.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "triples" "0.6.1.0.20251004.160701" "A flexible triple-based database for use in apps" '((seq "2.0") (emacs "28.1")) :commit "f2b8fe6d500e3543bece24381fd4ce34a171c860" :authors '(("Andrew Hyatt" . "ahyatt@gmail.com")) :maintainer '("Andrew Hyatt" . "ahyatt@gmail.com") :keywords '("triples" "kg" "data" "sqlite") :url "https://github.com/ahyatt/triples")
